# Copyright 2013-2022 Lawrence Livermore National Security, LLC and other
# Spack Project Developers. See the top-level COPYRIGHT file for details.
#
# SPDX-License-Identifier: (Apache-2.0 OR MIT)
from spack import *
from spack.pkg.builtin.boost import Boost
import os


class Hxtorch(WafPackage):
    """hxtorch --- a PyTorch-based toplevel for the BrainScaleS-2 neuromorphic hardware systems"""

    homepage = "https://github.com/electronicvisions/hxtorch"
    # This repo provides a waf binary used for the build below
    git      = "https://github.com/electronicvisions/pynn-brainscales.git"

    version('4.0-a3', branch='waf')

    # PPU compiler dependencies
    depends_on('oppulance@4.0-a3')

    # host software dependencies
    depends_on('bitsery', type=('build', 'link', 'run'))
    depends_on('binutils+gold+ld+plugins', type=('build', 'link', 'run')) # specialize
    depends_on(Boost.with_default_variants)
    depends_on('boost@1.69.0: +graph+icu+mpi+python+numpy+coroutine+context cxxstd=17', type=('build', 'link', 'run')) # specialize boost (non-clingo, type=('build', 'link', 'run'))
    depends_on('cereal', type=('build', 'link', 'run'))
    depends_on('cppcheck', type=('build', 'link', 'run'))
    depends_on('doxygen+graphviz', type=('build', 'link', 'run'))
    depends_on('genpybind@ebrains', type=('build', 'link', 'run'))
    depends_on('gflags', type=('build', 'link', 'run'))
    depends_on('googletest@1.11.0: +gmock', type=('build', 'link', 'run')) # variadic templates needed
    depends_on('intel-tbb', type=('build', 'link', 'run'))  # ppu gdbserver
    depends_on('libelf', type=('build', 'link', 'run'))
    depends_on('liblockfile', type=('build', 'link', 'run'))
    depends_on('llvm', type=('build', 'link', 'run'))
    depends_on('log4cxx', type=('build', 'link', 'run'))
    depends_on('pkgconfig', type=('build', 'link', 'run'))
    depends_on('python@3.7.0:', type=('build', 'link', 'run')) # BrainScaleS(-2, type=('build', 'link', 'run')) only supports Python >= 3.7
    depends_on('py-h5py', type=('build', 'link', 'run')) # PyNN tests need it
    depends_on('py-matplotlib', type=('build', 'link', 'run'))
    depends_on('py-nose', type=('build', 'link', 'run'))
    depends_on('py-numpy', type=('build', 'link', 'run'))
    depends_on('py-pybind11', type=('build', 'link', 'run'))
    depends_on('py-pybind11-stubgen', type=('build', 'link', 'run'))
    depends_on('py-pycodestyle', type=('build', 'link', 'run'))
    depends_on('py-pyelftools', type=('build', 'link', 'run'))
    depends_on('py-pylint', type=('build', 'link', 'run'))
    depends_on('py-torch', type=('build', 'link', 'run'))
    depends_on('py-pyyaml', type=('build', 'link', 'run'))
    depends_on('py-scipy', type=('build', 'link', 'run'))
    depends_on('py-sqlalchemy', type=('build', 'link', 'run'))
    depends_on('util-linux', type=('build', 'link', 'run'))
    depends_on('yaml-cpp+shared', type=('build', 'link', 'run'))
    extends('python')

    def setup_build_environment(self, env):
        """waf needs to find headers and libraries by itself (mostly `boost`
        tool, but also `gtest`); it also needs to run executables during
        configuration."""

        include = []
        for dep in self.spec.traverse(deptype='build'):
            query = self.spec[dep.name]
            try:
                if dep.name in ['pthreadpool', 'fxdiv']:
                    print('skipping {} in SPACK_INCLUDE_DIRS/CPATH/C{{,_PLUS}}_INCLUDE_PATH\n'.format(dep.name))
                    for d in query.headers.directories:
                        if os.path.exists(d):
                            env.remove_path('SPACK_INCLUDE_DIRS', d)
                    continue
                include.extend(query.headers.directories)
                print('headers (', dep.name, '):', query.headers.directories, "\n")
            except:
                pass

        library = []
        for dep in self.spec.traverse(deptype=('link', 'run')):
            query = self.spec[dep.name]
            # we probably should skip/remove the entries for the same
            # dependencies as above; however, ".libs" relies on the package
            # name being the same as the library name (e.g. approx.
            # libPACKAGE.so) so we are just moving googletest to the front
            if dep.name == 'googletest':
                for d in [self.spec['googletest'].prefix.lib64, self.spec['googletest'].prefix.lib]:
                    if os.path.exists(d):
                        print('moving to front of SPACK_{LINK,RPATH}_DIRS\n', d)
                        env.remove_path('SPACK_LINK_DIRS', d)
                        env.prepend_path('SPACK_LINK_DIRS', d)
                        env.remove_path('SPACK_RPATH_DIRS', d)
                        env.prepend_path('SPACK_RPATH_DIRS', d)
            try:
                print('libs (', dep.name, '):', query.libs.directories, "\n")
                # extend at front
                library = query.libs.directories + library
            except:
                pass

        path = []
        for dep in self.spec.traverse(deptype=('build', 'link', 'run')):
            if dep.name in ['pthreadpool', 'fxdiv']:
                print('skipping {} for bin'.format(dep.name))
                continue
            query = self.spec[dep.name]
            try:
                path.append(query.prefix.bin)
                print('bin (', dep.name, '):', query.prefix.bin, "\n")
            except:
                pass

        # llvm might be built with ~shared_libs but still builds shared libs
        if not any('llvm' in lib for lib in library):
            print("libs: manually adding ", self.spec['llvm'].prefix.lib)
            library.append(self.spec['llvm'].prefix.lib)

        # explicitly add googletest library if it isn't found above;
        # adding in front of the list of libraries is a hack to circumvent
        # issues from vendoring packages like pthreadpool
        if not any('googletest' in lib for lib in library):
            if os.path.isdir(self.spec['googletest'].prefix.lib64):
                print("libs: manually inserting at front ", self.spec['googletest'].prefix.lib64)
                library.insert(0, self.spec['googletest'].prefix.lib64)
            if os.path.isdir(self.spec['googletest'].prefix.lib):
                print("libs: manually inserting at front ", self.spec['googletest'].prefix.lib)
                library.insert(0, self.spec['googletest'].prefix.lib)

        env.set('CPATH', ':'.join(include))
        env.set('C_INCLUDE_PATH', ':'.join(include))
        env.set('CPLUS_INCLUDE_PATH', ':'.join(include))
        env.set('LIBRARY_PATH', ':'.join(library))
        env.set('LD_LIBRARY_PATH', ':'.join(library))
        env.prepend_path('PATH', ':'.join(path))

    def setup_run_environment(self, env):
        env.prepend_path('PYTHONPATH', self.prefix.lib)

    # override configure step as we perform a project setup first
    def configure(self, spec, prefix):
        """Setup and configure the project."""

        self.waf('setup', '--repo-db-url=https://github.com/electronicvisions/projects', '--without-munge',
            '--project=hxtorch@ebrains-' + str(spec.version),
            '--project=haldls@ebrains-' + str(spec.version),
            '--project=grenade@ebrains-' + str(spec.version),
            '--project=code-format@ebrains-' + str(spec.version),
            '--project=logger@ebrains-' + str(spec.version),
            '--project=halco@ebrains-' + str(spec.version),
            '--project=hate@ebrains-' + str(spec.version),
            '--project=fisch@ebrains-' + str(spec.version),
            '--project=ztl@ebrains-' + str(spec.version),
            '--project=hxcomm@ebrains-' + str(spec.version),
            '--project=rant@ebrains-' + str(spec.version),
            '--project=pywrap@ebrains-' + str(spec.version),
            '--project=lib-boost-patches@ebrains-' + str(spec.version),
            '--project=sctrltp@ebrains-' + str(spec.version),
            '--project=hwdb@ebrains-' + str(spec.version),
            '--project=visions-slurm@ebrains-' + str(spec.version),
            '--project=flange@ebrains-' + str(spec.version),
            '--project=lib-rcf@ebrains-' + str(spec.version),
            '--project=bss-hw-params@ebrains-' + str(spec.version),
            '--project=libnux@ebrains-' + str(spec.version)
         )

        args = ['--prefix={0}'.format(self.prefix)]
        args += self.configure_args()

        self.waf('configure', '--build-profile=release', *args)

    def build_args(self):
        args = ['--keep', '--test-execnone', '-v']

        return args
